import java.util.*;
class ClauseSet {
	ArrayList<Clause> clauses;
	int size;
	
	public ClauseSet() {
		clauses = new ArrayList<Clause>();
		size = 0;
	}

	public String toString() {
		return this.clauses.toString();
	}
	
	Clause get(int n) {
		return this.clauses.get(n);
	}

	boolean contains(Clause C) {
		for (Clause D : this.clauses) {
			if (C.equals(D)) {
				return true;
			}
		}
		return false;
	}
	
	void add(Clause C) {
		if (!this.contains(C)) {
			this.clauses.add(C.duplicate());
			this.size++;
		}
	}
	
	void remove(Clause C) {
		for (Clause D : this.clauses) {
			if (C.equals(D)) {
				this.clauses.remove(D);
				this.size--;
				break;
			}
		}
	}
	
	boolean subset(ClauseSet X) {
		for (Clause C : this.clauses) {
			if (!X.contains(C)) {
				return false;
			}
		}
		return true;
	}
	
	boolean equals(ClauseSet X) {
		if (this.subset(X) && X.subset(this)) {
			return true;
		}
		return false;
	}
	
	ClauseSet duplicate() {
		ClauseSet X = new ClauseSet();
		for (int i = 0; i < this.size; i++) {
			X.add(this.get(i));
		}
		return X;
	}
	
	boolean isEmpty() {
		if (this.clauses.isEmpty()) {
			return true;
		}
		return false;
	}

	ClauseSet union(ClauseSet X) {
		ClauseSet Y = new ClauseSet();
		for (Clause C : this.clauses) {
			Y.add(C);
		}
		for (Clause C : X.clauses) {
			Y.add(C);
		}
		return Y;
	}
	
	ClauseSet setminus(ClauseSet X) {
		ClauseSet Y = new ClauseSet();
		for (Clause C : this.clauses) {
			if (!X.contains(C)) {
				Y.add(C);
			}
		}
		return Y;
	}

	ClauseSet resolution(ClauseSet U) {
		ClauseSet X = new ClauseSet();
		ClauseSet Y = new ClauseSet();
		ClauseSet Z = new ClauseSet();
		X = this.duplicate();
		Y = U.duplicate();
		for (int i = 0; i < this.size; i++) {
			for (int j = 0; j < U.size; j++) {
				Z = Z.union(X.get(i).resolvents(Y.get(j)));
			}
		}
		return Z.setminus(X).setminus(Y);
	}
	
	void refute() {
		ClauseSet X = new ClauseSet();
		ClauseSet Y = new ClauseSet();
		X = this.duplicate();
		Y = this.duplicate();
		int n = 0;
		System.out.println("R^{" + n + "}(X) = " + X);
		while (!Y.hasSquare()) {
			n++;
			Y = X.resolution(Y);
			System.out.println("");
			System.out.println(n + "th resolvent set = " + Y);
			X = X.union(Y);
			System.out.println("");
			System.out.println("R^{" + n + "}(X) = " + X);
			if (Y.isEmpty()) {
				break;
			}
		}
	}
	
	boolean hasSquare() {
		for (Clause C : this.clauses) {
			if (C.isEmpty()) {
				return true;
			}
		}
		return false;
	}
}
