import java.util.*;
class TermSet {
	ArrayList<Term> terms;
	int size;
	
	public TermSet() {
		terms = new ArrayList<Term>();
		size = 0;
	}
	
	public String toString() {
		return this.terms.toString();
	}
	
	void add(Term t) {
		int n = 1;
		for (int i = 0; i < this.size; i++) {
			if (t.equals(this.terms.get(i))) {
				n = -1;
				break;
			}
		}
		if (n == 1) {
			this.terms.add(t.duplicate());
			this.size += 1;
		}
	}
	
	void remove(Term t) {
		for (int i = 0; i < this.size; i++) {
			if (this.terms.get(i).equals(t)) {
				this.terms.remove(this.terms.get(i));
				this.size -= 1;
			}
		}
	}

	Term get(int n) {
		return this.terms.get(n);
	}
	
	boolean contains(Term t) {
		for (Term tt : this.terms) {
			if (tt.equals(t)) {
				return true;
			}
		}
		return false;
	}
	
	boolean subset(TermSet tset) {
		for (int i = 0; i < this.size; i++) {
			if (!tset.contains(this.terms.get(i))) {
				return false;
			}
		}
		return true;
	}
	
	boolean equals(TermSet tset) {
		if (this.subset(tset) && tset.subset(this)) {
			return true;
		}
		return false;
	}
	
	TermSet union(TermSet tset) {
		TermSet ttset = new TermSet();
		for (int i = 0; i < this.size; i++) {
			ttset.add(this.terms.get(i));
		}
		for (int i = 0; i < tset.size; i++) {
			ttset.add(tset.terms.get(i));
		}
		return ttset;
	}

	boolean isEmpty() {
		if (this.size == 0) {
			return true;
		}
		return false;
	}
	
	TermSet duplicate() {
		TermSet tset = new TermSet();
		for (int i = 0; i < this.size; i++) {
			tset.add(this.terms.get(i).duplicate());
		}
		return tset;
	}
	
	TermSet substitute(Sbst sbst) {
		TermSet tset = new TermSet();
		for (int i = 0; i < this.size; i++) {
			tset.add(this.terms.get(i).substitute(sbst));
		}
		return tset;
	}
	
	TermSet substituteAtOnce(SbstSet sset) {
		TermSet tset = new TermSet();
		tset = this.duplicate();
		for (Sbst sbst : sset.sbsts) {
			tset = tset.substitute(sbst);
		}
		return tset;
	}
	
	boolean agreeModuloArguments() {
		for (int i = 0; i < this.terms.size(); i++) {
			if (!this.terms.get(0).agreeModuloArguments(this.terms.get(i))) {
				return false;
			}
		}
		return true;
	}
	
	//svWԂD
	TermSet disagreementSet() {
		TermSet tset = new TermSet();
		tset = this.duplicate();
		TermSet dset = new TermSet();
		if (this.size == 0) {
			return dset;
		}
		int n = 1;
		for (int i = 0; i < this.size; i++) {
			if (!this.get(0).equals(this.get(i))) {
				n = 0;
				break;
			}
		}
		if (n == 1) {
			return dset;
		}
		if (!tset.agreeModuloArguments()) {
			return tset;
		}
		for (int i = 0; i < tset.terms.get(0).arguments.length; i++) {
			TermSet tset2 = new TermSet();
			for (int j = 0; j < tset.size; j++) {
				tset2.add(tset.terms.get(j).arguments[i]);
			}
			dset = tset2.disagreementSet();
			if (!dset.isEmpty()) {
				break;
			}
		}
		return dset;
	}

	//svWԂ\bh̃eXg
	TermSet disagreementSetTest() {
		TermSet tset = new TermSet();
		tset = this.duplicate();
		TermSet dset = new TermSet();
		if (this.size == 0) {
			System.out.println("debug print: Empty, to begin with!");
			return dset;
		}
		int n = 1;
		for (int i = 0; i < this.size; i++) {
			if (!this.get(0).equals(this.get(i))) {
				n = 0;
				break;
			}
		}
		if (n == 1) {
			System.out.println("debug print: No disagreement arise!");
			return dset;
		}
		if (!tset.agreeModuloArguments()) {
			System.out.println("debug print: This doesn't agree even modulo arguments!");
			return tset;
		}
		for (int i = 0; i < tset.terms.get(0).arguments.length; i++) {
			TermSet tset2 = new TermSet();
			for (int j = 0; j < tset.size; j++) {
				tset2.add(tset.terms.get(j).arguments[i]);
				System.out.println("debug print: tset2 = " + tset2);
			}
			dset = tset2.disagreementSetTest();
			if (!dset.isEmpty()) {
				break;
			}
		}
		return dset;
	}
	
	TermSet freeVariableSet() {
		TermSet tset1 = new TermSet();
		tset1 = this.duplicate();
		TermSet tset2 = new TermSet();
		for (int i = 0; i < tset1.size; i++) {
			tset2 = tset2.union(tset1.terms.get(i).freeVariableSet());
		}
		return tset2;
	}

	SbstSet variableSeparation(TermSet tset) {
		TermSet fvset1 = new TermSet();
		fvset1 = this.freeVariableSet();
		TermSet fvset2 = new TermSet();
		fvset2 = tset.freeVariableSet();
		TermSet is = new TermSet();
		for (int i = 0; i < fvset1.size; i++) {
			if (fvset2.contains(fvset1.terms.get(i))) {
				is.add(fvset1.terms.get(i));
			}
		}
		SbstSet sset = new SbstSet();
		if (!is.isEmpty()) {
			int m = 0;
			for (int i = 0; i < fvset1.size; i++) {
				if (m < fvset1.terms.get(i).index) {
					m = fvset1.terms.get(i).index;
				}
			}
			for (int i = 0; i < fvset2.size; i++) {
				if (m < fvset2.terms.get(i).index) {
					m = fvset2.terms.get(i).index;
				}
			}
			m++;
			for (int i = 0; i < is.size; i++) {
				int x = is.terms.get(i).index;
				Sbst sbst = new Sbst();
				Term v = new Term(0, x, 0);
				Term t = new Term(0, m, 0);
				sbst.variable = v;
				sbst.term = t;
				sset.add(sbst);
				m++;
			}
		}
		return sset;
	}

	//PꉻASYD
	//̏W炻̍ŔēqimgujD
	//Pꉻ\łȂnullԂD
	SbstSet unifier() {
		TermSet tset = this.duplicate();
		TermSet dset = new TermSet();
		dset = tset.disagreementSet();
		SbstSet mgu = new SbstSet();
		while (!dset.isEmpty()) {
			int n = -1;
			for (int i = 0; i < dset.size; i++) {
				if (dset.get(i).kind == 0) {
					n = i;
					break;
				}
			}
			if (n == -1) {
				return null;
			}
			for (int i = 0; i < dset.size; i++) {
				if (i != n) {
					if (!dset.get(n).equals(dset.get(i))) {
						if (dset.get(n).subTerm(dset.get(i))) {
							return null;
						}
					}
				}
			}
			Sbst sbst = new Sbst();
			if (n == 0) {
				sbst.variable = dset.get(0);
				sbst.term = dset.get(1);
			} else {
				sbst.variable = dset.get(n);
				sbst.term = dset.get(0);
			}
			tset = tset.substitute(sbst);
			SbstSet sset = new SbstSet();
			sset.add(sbst);
			mgu = mgu.composite(sset);
			dset = tset.disagreementSet();
		}
		return mgu;
	}

	//ŔĒPꉻqԂ\bh̃eXg
	SbstSet unifierTest() {
		TermSet tset = this.duplicate();
		TermSet dset = new TermSet();
		dset = tset.disagreementSet();
		System.out.println("debug print: dset = " + dset);
		SbstSet mgu = new SbstSet();
		while (!dset.isEmpty()) {
			int n = -1;
			for (int i = 0; i < dset.size; i++) {
				if (dset.get(i).kind == 0) {
					n = i;
					break;
				}
			}
			if (n == -1) {
				System.out.println("debug print: No variable contained!");
				return null;
			}
			for (int i = 0; i < dset.size; i++) {
				if (i != n) {
					if (!dset.get(n).equals(dset.get(i))) {
						if (dset.get(n).subTerm(dset.get(i))) {
							System.out.println("debug print: " + dset.get(n) + " is a subterm of " + dset.get(i));
							return null;
						}
					}
				}
			}
			Sbst sbst = new Sbst();
			if (n == 0) {
				sbst.variable = dset.get(0);
				sbst.term = dset.get(1);
			} else {
				sbst.variable = dset.get(n);
				sbst.term = dset.get(0);
			}
			System.out.println("debug print: sbst = " + sbst);
			tset = tset.substitute(sbst);
			System.out.println("debug print: tset = " + tset);
			SbstSet sset = new SbstSet();
			sset.add(sbst);
			mgu = mgu.composite(sset);
			System.out.println("debug print: mgu = " + mgu);
			dset = tset.disagreementSet();
			System.out.println("debug print: dset = " + dset);
		}
		return mgu;
	}	
	boolean unifiable() {
		if (this.unifier() == null) {
			return false;
		}
		return true;
	}
	
	public static void main(String[] args) {
		Term v1 = new Term(0, 1, 0);
		Term v2 = new Term(0, 2, 0);
		Term v3 = new Term(0, 3, 0);
		Term c1 = new Term(1, 1, 0);
		Term c2 = new Term(1, 2, 0);
		Term f1 = new Term(1, 1, 2);
		Term f2 = new Term(1, 1, 2);
		Term f3 = new Term(1, 1, 2);
		Term f4 = new Term(1, 1, 2);
		Term f5 = new Term(1, 1, 2);
		f1.arguments[0] = v1;
		f1.arguments[1] = c2;
		f2.arguments[0] = v1;
		f2.arguments[1] = v3;
		f3.arguments[0] = v1;
		f3.arguments[1] = f1;
		f4.arguments[0] = c2;
		f4.arguments[1] = f2;
		f5.arguments[0] = f1;
		f5.arguments[1] = f2;
		Sbst sbst1 = new Sbst();
		Sbst sbst2 = new Sbst();
		Sbst sbst3 = new Sbst();
		Sbst sbst4 = new Sbst();
		sbst1.term = c1;
		sbst1.variable = v1;
		sbst2.term = c2;
		sbst2.variable = v2;
		sbst3.term = f2;
		sbst3.variable = v1;
		sbst4.term = c1;
		sbst4.variable = v3;
		SbstSet sset1 = new SbstSet();
		TermSet tset1 = new TermSet();
		TermSet tset2 = new TermSet();
		TermSet tset3 = new TermSet();
		tset1.add(f3);
		tset1.add(f4);
		tset1.add(f5);
		tset2.add(v2);
		tset2.add(f2);
		sset1 = tset2.variableSeparation(tset1);
		System.out.println(tset1);
		System.out.println(tset2);
		System.out.println(sset1);
		System.out.println(tset1.substituteAtOnce(sset1));
	}
}
